from singer import metadata
import singer
from ankit_db.sync_strategies import log_based
from ankit_db.sync_strategies import full_table

LOGGER = singer.get_logger()

def clear_state_on_replication_change(stream, state):
    md_map = metadata.to_map(stream['metadata'])
    tap_stream_id = stream['tap_stream_id']

    # replication method changed
    current_replication_method = metadata.get(md_map, (), 'replication-method')
    last_replication_method = singer.get_bookmark(state, tap_stream_id, 'last_replication_method')
    if last_replication_method is not None and (current_replication_method != last_replication_method):
        log_msg = 'Replication method changed from %s to %s, will re-replicate entire table %s'
        LOGGER.info(log_msg, last_replication_method, current_replication_method, tap_stream_id)
        state = singer.reset_stream(state, tap_stream_id)

    state = singer.write_bookmark(state, tap_stream_id, 'last_replication_method', current_replication_method)

    return state

def sync_stream(config, state, stream):
    table_name = stream['tap_stream_id']

    md_map = metadata.to_map(stream['metadata'])

    replication_method = metadata.get(md_map, (), 'replication-method')
    key_properties = metadata.get(md_map, (), 'table-key-properties')

    rmSet = {'LOG_BASED'}
    if replication_method not in rmSet:
        LOGGER.info('Unknown replication method: %s ', replication_method)
        replication_method = 'LOG_BASED'
        LOGGER.info('version 9 Setting default replication method: %s ', replication_method)

    # Update replication method if it is not equal to prev replication method
    state = clear_state_on_replication_change(stream, state)
    
    state = singer.set_currently_syncing(state, table_name)
    singer.write_state(state)

    singer.write_message(singer.SchemaMessage(
        stream=table_name,
        schema=stream['schema'],
        key_properties=key_properties))

    rows_saved = 0
    if replication_method == 'FULL_TABLE':
        LOGGER.info("Syncing full table for stream: %s", table_name)
        rows_saved += full_table.sync(config, state, stream)
    elif replication_method == 'LOG_BASED':
        LOGGER.info("Syncing log based for stream: %s", table_name)

        if log_based.has_stream_aged_out(state, table_name):
            LOGGER.info("Clearing state because stream has aged out")
            state.get('bookmarks', {}).pop(table_name)

        if not singer.get_bookmark(state, table_name, 'initial_full_table_complete'):
            msg = 'Must complete full table sync before replicating from dynamodb streams for %s'
            LOGGER.info(msg, table_name)

            state = log_based.get_initial_bookmarks(config, state, table_name)
            singer.write_state(state)

            rows_saved += full_table.sync(config, state, stream)

        rows_saved += log_based.sync(config, state, stream)
    else:
        LOGGER.info('Unknown replication method: %s for stream: %s', replication_method, table_name)

    state = singer.write_bookmark(state, table_name, 'success_timestamp', singer.utils.strftime(singer.utils.now()))
    singer.write_state(state)

    return rows_saved



from singer import metadata
import singer
import backoff
from botocore.exceptions import ClientError, ConnectTimeoutError, ReadTimeoutError
from ankit_db import dynamodb

LOGGER = singer.get_logger()

def discover_table_schema(client, table_name):
    '''
    For each of the tables in the dynamodb, create the table schema for the catalog
    '''
    try:
        table_info = client.describe_table(TableName=table_name).get('Table', {})
    except ClientError:
        LOGGER.info('Access to table %s was denied, skipping', table_name)
        return None

    # write stream metadata
    mdata = {}
    key_props = [key_schema.get('AttributeName') for key_schema in table_info.get('KeySchema', [])]
    mdata = metadata.write(mdata, (), 'table-key-properties', key_props)
    if table_info.get('ItemCount'):
        mdata = metadata.write(mdata, (), 'row-count', table_info['ItemCount'])

    return {
        'table_name': table_name,
        'stream': table_name,
        'tap_stream_id': table_name,
        'metadata': metadata.to_list(mdata),
        'schema': {
            'type': 'object'
        }
    }

# Backoff for both ReadTimeout and ConnectTimeout error for 5 times
@backoff.on_exception(backoff.expo,
                    (ReadTimeoutError, ConnectTimeoutError),
                    max_tries=5,
                    factor=2)
def discover_streams(config):
    '''
    Get the list of the tables and create the table schema for each of them
    '''
    client = dynamodb.get_client(config)

    try:
        response = client.list_tables()
    except ClientError:
        raise Exception("Authorization to AWS failed. Please ensure the role and policy are configured correctly on your AWS account.")

    table_list = response.get('TableNames')

    while response.get('LastEvaluatedTableName') is not None:
        response = client.list_tables(ExclusiveStartTableName=response.get('LastEvaluatedTableName'))
        table_list += response.get('TableNames')
    streams = [x for x in
               (discover_table_schema(client, table) for table in table_list)
               if x is not None]

    LOGGER.info("getting the streams here : %s", streams)
    return streams



import base64
import decimal
from boto3.dynamodb.types import TypeDeserializer

# Custom context to control how decimals are deserialized
# Precision = 100 because that's the max for `singer.decimal` elsewhere, still
# trapping Rounding errors because that is a true error.
trapped_signals = [decimal.Clamped, decimal.Overflow, decimal.Inexact, decimal.Rounded, decimal.Underflow]
SINGER_CONTEXT = decimal.Context(Emin=-128, Emax=126, prec=100,
                                 traps=trapped_signals)

class Deserializer(TypeDeserializer):
    '''
    This class inherits from boto3.dynamodb.types.TypeDeserializer
    https://boto3.amazonaws.com/v1/documentation/api/latest/_modules/boto3/dynamodb/types.html
    By extending TypeDeserializer we get a lot of functionality for
    free (ie handling nested data) while allowing us to redefine how to
    handle binary data and sets
    '''

    def deserialize_item(self, item):
        return self.deserialize({'M': item})

    def _deserialize_b(self, value):
        '''
        Deserializes binary data as a base64 encoded string because that's how
        the aws cli returns binary data
        '''
        return base64.b64encode(value).decode('utf-8')

    def _deserialize_n(self, value):
        '''
        Deserializes sets as lists to allow JSON encoding
        '''
        return SINGER_CONTEXT.create_decimal(value)

    def _deserialize_ns(self, value):
        '''
        Deserializes sets as lists to allow JSON encoding
        '''
        return list(map(self._deserialize_n, value))

    def _deserialize_ss(self, value):
        '''
        Deserializes sets as lists to allow JSON encoding
        '''
        return list(map(self._deserialize_s, value))

    def _deserialize_bs(self, value):
        '''
        Deserializes sets as lists to allow JSON encoding
        '''
        return list(map(self._deserialize_b, value))

    def _apply_projection(self, record, breadcrumb, output):
        """
        The LOG_BASED replication method uses the get_records method which gets all the records by default.
        In case of projection expression, parse the response retrieved from the API and filter the output
        based on projection expressions
        """
        if len(breadcrumb) == 1:
            if '[' in breadcrumb[0]:
                breadcrumb_key = breadcrumb[0].split('[')[0]
                index = int(breadcrumb[0].split('[')[1].split(']')[0])
                if output.get(breadcrumb_key):
                    # only prepare output if the list field contains data at that index position in record
                    if len(record.get(breadcrumb_key)) > index:
                        output[breadcrumb_key].append(record[breadcrumb_key][index])
                else:
                    output[breadcrumb_key] = []
                    # only prepare output if the list field contains data at that index position in record
                    if record.get(breadcrumb_key) and len(record.get(breadcrumb_key)) > index:
                        output[breadcrumb_key].append(record[breadcrumb_key][index])
            else:
                output[breadcrumb[0]] = record.get(breadcrumb[0])
        else:
            if '[' in breadcrumb[0]:
                breadcrumb_key = breadcrumb[0].split('[')[0]
                index = int(breadcrumb[0].split('[')[1].split(']')[0])
                if not output.get(breadcrumb_key):
                    output[breadcrumb_key] = [{}]
                # only prepare output if the list field contains data at that index position in record
                if record.get(breadcrumb_key) and len(record.get(breadcrumb_key)) > index:
                    self._apply_projection(record[breadcrumb_key][index], breadcrumb[1:], output[breadcrumb_key][0])
            else:
                if output.get(breadcrumb[0]) is None:
                    output[breadcrumb[0]] = {}
                # keep empty dict if the data is not found in the record
                if record.get(breadcrumb[0]):
                    self._apply_projection(record.get(breadcrumb[0], {}), breadcrumb[1:], output[breadcrumb[0]])

    def apply_projection(self, record, projections):
        output = {}

        for breadcrumb in projections:
            self._apply_projection(record, breadcrumb, output)

        return output



import json
import backoff
import boto3
import singer

from botocore.credentials import (
    AssumeRoleCredentialFetcher,
    CredentialResolver,
    DeferredRefreshableCredentials,
    JSONFileCache
)
from botocore.session import Session
from botocore.config import Config
from botocore.exceptions import ClientError, ConnectTimeoutError, ReadTimeoutError

LOGGER = singer.get_logger()
REQUEST_TIMEOUT = 300

def retry_pattern():
    '''
    Retry logic for the ClientError
    '''
    return backoff.on_exception(backoff.expo,
                                (ClientError, ConnectTimeoutError, ReadTimeoutError),
                                max_tries=5,
                                on_backoff=log_backoff_attempt,
                                factor=10)


def log_backoff_attempt(details):
    '''
    Log the retry attempt with the number of tries
    '''
    LOGGER.info("Error detected communicating with Amazon, triggering backoff: %d try", details.get("tries"))


class AssumeRoleProvider():
    METHOD = 'assume-role'

    def __init__(self, fetcher):
        self._fetcher = fetcher

    def load(self):
        return DeferredRefreshableCredentials(
            self._fetcher.fetch_credentials,
            self.METHOD
        )


@retry_pattern()
def setup_aws_client(config):
    """
    Setup the aws session for making the API calls
    """
    if 'role_name' in config:
        role_arn = "arn:aws:iam::{}:role/{}".format(config['account_id'].replace('-', ''),
                                                    config['role_name'])

        session = Session()
        fetcher = AssumeRoleCredentialFetcher(
            session.create_client,
            session.get_credentials(),
            role_arn,
            extra_args={
                'DurationSeconds': 3600,
                'RoleSessionName': 'TapDynamodDB',
                'ExternalId': config['external_id']
            },
            cache=JSONFileCache()
        )

        refreshable_session = Session()
        refreshable_session.register_component(
            'credential_provider',
            CredentialResolver([AssumeRoleProvider(fetcher)])
        )

        LOGGER.info("Attempting to assume_role on RoleArn: %s", role_arn)
        boto3.setup_default_session(botocore_session=refreshable_session)

def get_request_timeout(config):
    # if request_timeout is other than 0,"0" or "" then use request_timeout
    request_timeout = config.get('request_timeout')
    if request_timeout and float(request_timeout):
        request_timeout = float(request_timeout)
    else: # If value is 0,"0" or "" then set default to 300 seconds.
        request_timeout = REQUEST_TIMEOUT
    return request_timeout

def get_client(config):
    """
    Client for FULL_TABLE and running discover mode.
    """
    # get the request_timeout
    request_timeout = get_request_timeout(config)
    # add the request_timeout in both connect_timeout as well as read_timeout
    timeout_config = Config(connect_timeout=request_timeout, read_timeout=request_timeout)
    if config.get('use_local_dynamo'):
        return boto3.client('dynamodb',
                            endpoint_url='http://localhost:8000',
                            region_name=config['region_name'],
                            config=timeout_config   # pass the config to add the request_timeout
                            )
    return boto3.client('dynamodb', config['region_name'],
                        config=timeout_config   # pass the config to add the request_timeout
                        )

def get_stream_client(config):
    """
    Streams client for the LOG_BASED sync.
    """
    # get the request_timeout
    request_timeout = get_request_timeout(config)
    # add the request_timeout in both connect_timeout as well as read_timeout
    timeout_config = Config(connect_timeout=request_timeout,  read_timeout=request_timeout)
    if config.get('use_local_dynamo'):
        return boto3.client('dynamodbstreams',
                            endpoint_url='http://localhost:8000',
                            region_name=config['region_name'],
                            config=timeout_config   # pass the config to add the request_timeout
                            )
    return boto3.client('dynamodbstreams',
                        region_name=config['region_name'],
                        config=timeout_config   # pass the config to add the request_timeout
                        )

def decode_expression(expression):
    '''Convert the string into JSON object and raise an exception if invalid JSON format'''
    try:
        return json.loads(expression)
    except json.decoder.JSONDecodeError:
        raise Exception("Invalid JSON format. The expression attributes should contain a valid JSON format.")



import json
import sys
import time

from terminaltables import AsciiTable
import singer
from singer import metadata
from ankit_db.discover import discover_streams
from ankit_db.dynamodb import setup_aws_client
from ankit_db.sync import sync_stream


LOGGER = singer.get_logger()

REQUIRED_CONFIG_KEYS = ["region_name"]

def do_discover(config):
    '''
    Run the discovery mode for each streams
    '''
    LOGGER.info("Starting discover version updated")
    streams = discover_streams(config)
    if not streams:
        raise Exception("No streams found")
    catalog = {"streams": streams}
    json.dump(catalog, sys.stdout, indent=2)
    LOGGER.info("Finished discover")

def stream_is_selected(mdata):
    '''
    Get the streams which are selected from the catalog
    '''
    return mdata.get((), {}).get('selected', False)

def do_sync(config, catalog, state):
    '''
    Run the sync mode for each streams
    '''
    LOGGER.info('Starting sync.')

    counts = {}
    sync_times = {}
    for stream in catalog['streams']:
        start_time = time.time()
        stream_name = stream['tap_stream_id']
        mdata = metadata.to_map(stream['metadata'])
        if not stream_is_selected(mdata):
            LOGGER.info("%s: Skipping - not selected", stream_name)
            continue

        singer.write_state(state)
        key_properties = metadata.get(mdata, (), 'table-key-properties')
        singer.write_schema(stream_name, stream['schema'], key_properties)

        LOGGER.info("%s: Starting sync", stream_name)
        counts[stream_name] = sync_stream(config, state, stream)
        sync_times[stream_name] = time.time() - start_time
        LOGGER.info("%s: Completed sync (%s rows)", stream_name, counts[stream_name])

    LOGGER.info(get_sync_summary(catalog, counts, sync_times))
    LOGGER.info('Done syncing.')

def get_sync_summary(catalog, counts, times):
    '''
    Obtain the sync summary for all the stream tables
    '''
    headers = [['table name',
                'replication method',
                'total records',
                'write speed']]

    rows = []
    for stream_id, stream_count in counts.items():
        stream = [x for x in catalog['streams'] if x['tap_stream_id'] == stream_id][0]
        md_map = metadata.to_map(stream['metadata'])
        replication_method = metadata.get(md_map, (), 'replication-method')

        stream_time = times[stream_id]
        if stream_time == 0:
            stream_time = 0.000001
        row = [stream_id,
               replication_method,
               '{} records'.format(stream_count),
               '{:.1f} records/second'.format(stream_count/stream_time)]
        rows.append(row)

    data = headers + rows
    table = AsciiTable(data, title='Sync Summary')

    return '\n\n' + table.table


@singer.utils.handle_top_exception(LOGGER)
def main():
    args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)
    config = args.config

    # TODO Is this the right way to do this? It seems bad
    if not config.get('use_local_dynamo'):
        setup_aws_client(config)

    if args.discover:
        do_discover(args.config)
    elif args.catalog:
        do_sync(config, args.catalog.to_dict(), args.state)


if __name__ == '__main__':
    main()
